/*

    File: testdisk.c

    Copyright (C) 1998-2002  Christophe GRENIER <grenier@cgsecurity.org>
  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 */

#include <stdarg.h>
#include <unistd.h>	/* geteuid */
#include <string.h>
#include <time.h>
#include <sys/time.h>
#include <sys/types.h>
#include <ctype.h>      /* toupper, tolower */
#include "types.h"
#include "common.h"
#include "testdisk.h"
#include "lang.h"
#include "intrface.h"
#include "godmode.h"
#include "fnctdsk.h"
#include "version.h"

int load_dump(const char*filename, unsigned char * buffer, const int size);

FILE *f_rapport=NULL;
int f_status=0;

void ecrit_rapport_string(const char *string,const int max_length)
{
  int i;
  for(i=0;(string[i]!='\0')&&(i<max_length);i++)
    ecrit_rapport("%c",string[i]);
}

int ecrit_rapport(const char *_format, ...)
{
  int res=0;
  if(f_rapport!=NULL)
  {
    va_list ap;
    va_start(ap,_format);
    res=vfprintf(f_rapport,_format,ap);
    va_end(ap);
    if(fflush(f_rapport))
    {
      f_status=1;
    }
  }
  return res;
}

int doprintf(const char *_format, ...)
{
  char res[800];
  va_list ap;
  va_start(ap,_format);
#ifdef DEBUG
  if(f_rapport!=NULL)
    vfprintf(f_rapport,_format,ap);
#endif
#ifdef DJGPP
  vsprintf(res,_format,ap);
#else
  vsnprintf(res,sizeof(res),_format,ap);
#endif
  va_end(ap);
  return addstr(res);
}

int wdoprintf(WINDOW *window, const char *_format, ...)
{
  char res[800];
  va_list ap;
  va_start(ap,_format);
#ifdef DEBUG
  if(f_rapport!=NULL)
    vfprintf(f_rapport,_format,ap);
#endif
#ifdef DJGPP
  vsprintf(res,_format,ap);
#else
  vsnprintf(res,sizeof(res),_format,ap);
#endif
  va_end(ap);
  return waddstr(window,res);
}

int ask_YN(WINDOW *window,const char* msg)
{
  char res;
  wdoprintf(window,msg);
  wrefresh(window);
  do
  {
    res=toupper(getch());
  } while((res!=c_NO)&(res!=c_YES));
  wdoprintf(window,"%c\n",res);
  return (int)(res==c_YES);
}

FILE* init_log(const char*filename,int argc, char**argv)
{
  FILE*f_file=fopen(filename,"a");
  if(f_file==NULL)
	printf(m_LOG_ERR);
  else
  {
	int i;
	time_t my_time=time(NULL);
	fprintf(f_file,"\n\n%s",ctime(&my_time));
	fprintf(f_file,m_CMDLINE);
	for(i=1;i<argc;i++)
	  fprintf(f_file," %s", argv[i]);
	fprintf(f_file,"\n");
  }
  return f_file;
}


int main( int argc, char **argv )
{
  int i;
  unsigned int x;
  int help=0, create_log=0,paranoid=0,debug=0, dump_ind=0;
  int fast_mode=0;
  int align=1;
  int do_analyse=1;
  int do_list=0;
  int test_recovery=0;
#ifdef TESTING
  srand(1);
#endif
  for(i=1;i<argc;i++)
  {
    for (x=0; x < strlen( argv[i] ); x++)
     argv[i][x] = tolower( argv[i][x] );
    if((strcmp(argv[i],"/test_recovery")==0) ||(strcmp(argv[i],"-test_recovery")==0))
    {
      test_recovery=1;
      do_list=1;
      create_log=1;
    }
    else if((strcmp(argv[i],"/dump")==0) || (strcmp(argv[i],"-dump")==0))
      dump_ind=1;
    else if((strcmp(argv[i],"/log")==0) ||(strcmp(argv[i],"-log")==0))
      create_log=1;
    else if((strcmp(argv[i],"/debug")==0) || (strcmp(argv[i],"-debug")==0))
    {
      debug++;
      create_log=1;
    }
    else if((strcmp(argv[i],"/list")==0) || (strcmp(argv[i],"-list")==0))
      do_list=1;
    else
      help=1;
  }
  printf(m_Copyright);
  if(help!=0)
  {
    printf(m_Usage);
    return 0;
  }
  if(create_log!=0)
  {
    f_rapport=init_log("testdisk.log",argc,argv);
    ecrit_rapport(m_Copyright);
#ifdef DJGPP
    ecrit_rapport("Dos version");
#elif defined(BSD)
    ecrit_rapport("BSD version");
#elif defined(LINUX)
    ecrit_rapport("Linux version");
#else
    ecrit_rapport("Undefined OS");
#endif
#ifdef COMPILE_BY
#ifdef COMPILE_HOST
#ifdef COMPILE_TIME
    ecrit_rapport(" (%s@%s, %s)",COMPILE_BY,COMPILE_HOST,COMPILE_TIME);
#endif
#endif
#endif
    ecrit_rapport("\n\n");
#ifdef DEBUG
    ecrit_rapport("Key down:      0x%x\n", KEY_DOWN);
    ecrit_rapport("Key up:        0x%x\n", KEY_UP);
    ecrit_rapport("Key left:      0x%x\n", KEY_LEFT);
    ecrit_rapport("Key right:     0x%x\n", KEY_RIGHT);
    ecrit_rapport("Key page down: 0x%x\n", KEY_NPAGE);
    ecrit_rapport("Key page up:   0x%x\n", KEY_PPAGE);
    ecrit_rapport("Key enter:     0x%x\n", KEY_ENTER);
#endif
  }
  printf("Please wait...\n");
  if(do_list!=0)
  {
    t_param_disk param_disk[NBR_DISK_MAX];
    int nbr_disk;
    for(i=0;i<NBR_DISK_MAX;i++)
    {
      param_disk[i]=(t_param_disk)MALLOC(sizeof(*param_disk[i]));
#ifndef DJGPP
      param_disk[i]->name=NULL;
#endif
    }
    nbr_disk=hd_parse(param_disk,debug,0);
#ifndef DJGPP
    if(nbr_disk==0 && geteuid()!=0)
    {
      printf("You need to be root to use TestDisk\n");
      return 1;
    }
#endif
    /* save disk parameters to rapport */
    printf("Hard disk list\n");
    ecrit_rapport("Hard disk list\n");
    for(i=0;i<nbr_disk;i++)
    {
      aff_disk_classic(param_disk[i]);
      aff_disk_rapport(param_disk[i]);
    }
    printf("\n");
    ecrit_rapport("\n");
    for(i=0;i<nbr_disk;i++)
    {
      interface_list(param_disk[i],debug,test_recovery);
      if(test_recovery!=0)
      {
	t_list_part list_part;
	t_list_part element;
	//dump_ind=0;
	list_part=search_part(param_disk[i],paranoid,debug,0,fast_mode,0);
	if(debug!=0)
	{
	  ecrit_rapport("search_part() results\n");
	  for(element=list_part;element!=NULL;element=element->next)
	    aff_part_rapport(param_disk[i],element->part);
	}
	delete_list_part(list_part);
      }
    }
    for(i=0;i<NBR_DISK_MAX;i++)
    {
#ifndef DJGPP
      if(param_disk[i]->name)    /* forget name */
	FREE(param_disk[i]->name);
#endif
      FREE(param_disk[i]);
    }
  }
  else
  {
    if(do_curses_testdisk(debug,paranoid,dump_ind,fast_mode,align,do_analyse))
    {
      printf("TestDisk need 25 lines to work.\nPlease enlarge the terminal and restart TestDisk.\n");
    }
  }
  if(f_rapport)
  {
    ecrit_rapport("TestDisk exited normally.\n");
    if(fclose(f_rapport))
      f_status=1;
  }
  if(f_status!=0)
  {
    printf("TestDisk: Log file corrupted!\n");
  } else
    printf("TestDisk exited normally.\n");
  return 0;
}

#ifdef DEBUG

#include "fat.h"

int do_test()
{
  FILE *file_in;
  unsigned char buffer[10*0x200];
  unsigned int taille;
  t_param_disk disk_car=(t_param_disk)MALLOC(sizeof(*disk_car));
  t_diskext partition=partition_new(0);
  initscr();
  noecho();
  keypad(stdscr, TRUE); // Need it to get arrow key
#ifndef DJGPP
  nonl(); //don't use for Dos version but enter will work with it... dilema
#endif
  crmode();
  //  intrflush(stdscr, FALSE);
  cbreak();

  file_in=fopen("BOOT_BIN.txt","rb");
  if(!file_in)
    return 1;
  taille=fread(buffer,1,10*0x200,file_in);
  fclose(file_in);
  if(memcmp(buffer,"0000 ",5)==0)
  {
    int pos_file;
    int pos_buffer2=0;
    unsigned char buffer2[3*0x200];
    for(pos_file=5;pos_file<taille;)
    {
      char string[3];
      if(buffer[pos_file]==' ')
	pos_file++;
      string[0]=buffer[pos_file];
      string[1]=buffer[pos_file+1];
      string[2]=0;
      printf("%02X",strtol(string,NULL,16));
      buffer2[pos_buffer2++]=strtol(string,NULL,16);
      if(pos_buffer2%16==0)
      {
	pos_file+=18;
	if(buffer[pos_file]==0xA || buffer[pos_file]==0xD)
	  pos_file++;
	if(buffer[pos_file]==0xA || buffer[pos_file]==0xD)
	  pos_file++;
	pos_file+=9;
      }
      else
	pos_file+=2;
    }
//    buffer2_size=pos_buffer2-1;
    memcpy(buffer,buffer2,3*0x200);
  }
  printf("do_test\n");
  aff_buffer(BUFFER_RESET,"Q");
  dump_fat_info(buffer,P_16FAT);
  test_FAT(disk_car,(struct fat_boot_sector *)&buffer, partition,1, 1);
  aff_buffer(BUFFER_DISPLAY,"Q",stdscr);
  return 0;
}

int load_dump(const char*filename, unsigned char * buffer, const int size)
{
  FILE *f_in;
  unsigned char ligne[200];
  int position=0;
  int lu;
  int etat=0;
  unsigned char val=0;
  int position_ligne=0;
  memset(buffer,0,size);
  if((f_in=fopen(filename,"r"))==NULL)
    return 1;
//  ecrit_rapport("load_dump %s ok\n",filename);
  while(!feof(f_in) && ((lu=fread(ligne,1,sizeof(ligne)-1,f_in))>0))
  {
    int i;
    ligne[sizeof(ligne)-1]=0;
//    ecrit_rapport("ligne=%s,lu=%d\n",ligne,lu);
    for(i=0;i<lu;i++)
    {
      unsigned char car=ligne[i];
//      ecrit_rapport("etat=%d, car=%02X(%c), val=%02X, position_ligne=%d, position=%04X\n",etat,car,car,val,position_ligne,position);
      switch(etat)
      {
	case 0:
	  if(car==':')
	    etat=1;
	  break;
	case 1:
	  if(car!=' ')
	  {
	    if(car>='0' && car<='9')
	      val=(val<<4)|(car-'0');
	    else
	      val=(val<<4)|(car-'A'+10);
	    etat=2;
	  }
	  break;
	case 2:
	  if(car>='0' && car<='9')
	    val=(val<<4)|(car-'0');
	  else
	    val=(val<<4)|(car-'A'+10);
	  buffer[position]=val;
	  position++;
	  position_ligne++;
	  if(position_ligne<0x10)
	    etat=1;
	  else
	    etat=3;
	  break;
	case 3:
	  if(car=='\n')
	  {
	    etat=0;
	    position_ligne=0;
	  }
	  break;
      }	// end switch
    } // end for lu
  } // end while
  fclose(f_in);
  return 0;
}
#endif
